/*********************  PARETO_evolve.h ***************************************
Header file of constants, definitions, data structure declarations, and function 
declarations.  
USERS SHOULD ONLY EDIT THE UserSettings.h and UserArrays.h header files.
DO NOT EDIT THIS FILE. DO NOT EDIT THIS FILE. etc.
This file is 'included' in most every other file of PARETO_EVOLVE.
********************************************************************        
    Editing History: 
    21 May 1998 - creation; J. Reynolds & M. Turley.
    25 Feb, 7 & 10 Sept 2005 - documentation revision, argument list restructuring,
                                addition of elitism functions; R. Komuro. 
    26 Feb 2006 - documentation expansion, removal of components requiring
                  user editing to UserSettings.h; Joel H. Reynolds, joel_reynolds@fws.gov
*************************************************************************/
#ifndef _pareto_evolve_h
#define _pareto_evolve_h 1

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include "UserSettings.h"
  
/*******************************************************************/
/******* Do not edit the following entries. *****************/                                
                                

#define DEBUGFLAG 0  /* 0 - suppresses printing current code point on screen*/
                     /* 1 - print current location of execution, for debugging */        


           
           /* DEFINITIONS */
#define within(a,b,c) ((((a)>=(b))&&((a)<=(c))) ? 1:0)

static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)
        
/********************************************************************/      
        /* STRUCT DEFINITIONS */

/* For storing a simulation's Parameterization values. */
struct param {          
        double  par[NUMPARAMS];
        };

/* For storing a simulation's Criteria Results. */
struct criteria_res {           
        double  crit[NUMCRITERIA];
        };

/* Element of a linked list, stores a simulation's info. */
struct siminfo {                
        struct siminfo *next;

        struct param paramvalues;       /* array holding parameterization */
        struct criteria_res cresults;   /* array holding criteria results */
        struct criteria_res assess_res; /* array holding assessed criteria */

        int NDrank;              /* NonDominated Ranking. Initially, NDrank=0 */
        double niche_count;      /* Number of neighbors, included self.*/
        double fitness;          /* fitness of simulation */
        };

/* Element of a linked list, stores a simulation on the current Pareto 
Frontier.*/
struct psiminfo {       
        struct psiminfo *next;
        struct param paramvalues;
        struct criteria_res cresults;
        };

/* Element of a linked list, stores a GROUP of elements in the Pareto 
Frontier. If BINARY error measures are used, a GROUP = all simulations 
giving the identical assessment vector of 0's & 1's. If CONTINUOUS error 
measures are used, each simulation is its own GROUP.*/
struct pset {
        struct pset *next;

        struct criteria_res assess;  /* assessment vector identifying group*/
        struct psiminfo *psim;       /* ptr to list of simulations in 
                                        group*/
        };


/**********************************************************/
/***************** FUNCTION DECLARATIONS ******************/
/* See function code for full description. Major functions
   briefly described below.*/
/*********** functions in PARETO_update.c ******************/

/* Reads in last round of model simulation results, updates
   Pareto Frontier. */
void evaluation(struct pset **ptr_to_Frontier_first_location, 
                struct siminfo **ptr_to_head_location, FILE *input_ptr, 
                int *pareto_flag_ptr, double *total_fitness_ptr, int generation);

void read_input(int *erc_ptr,FILE *input_ptr, 
                struct siminfo *latest_sim_ptr,int *pop_ptr);
                
/* Creates a simulation's assessment vector from its criteria results.*/
void criteria_assess(struct siminfo *latest_sim_ptr);

/* Adds simulation results to list of already read-in simulation results.*/
void ordered_add_to_current_gen(struct siminfo *latest_sim_ptr, 
                        struct siminfo **ptr_to_head_location);


void insert(struct siminfo *ptr_to_prior_item, struct siminfo *ptr_to_following_item, 
             struct siminfo *latest_sim_ptr);

/* Compares two simulations' assessment vectors. */
int tie_test(struct siminfo *latest_sim_ptr, struct siminfo *comparison_sim_ptr, int crit);

int compare( double left_res, double right_res);

/* Calculates each simulation's nondominated ranking and niche_count. */
void assign_fitness(struct siminfo *current_gen_first_ptr, double *total_fitness);

double sharing(double dist);

void non_dominated_ranking(struct siminfo *current_sim, struct siminfo *comparison_sim);

/* Calculates distance between two simulations in the criteria space.*/
void criteria_distance(struct siminfo *current_sim, struct siminfo *comparison_sim);

int crit_comp(double Left, double Right);

void share_fitness(struct siminfo* current_gen_first_ptr, double max_niche_count[POP_SIZE],
                double *total_fitness);

 /* Updates ('Historical') Pareto Frontier as revealed by the previous 
 generations of the search to include the current generation's simulation 
 results.*/
void Pareto_Frontier_update(struct siminfo *current_gen_first_ptr,
        struct pset **ptr_to_Frontier_first_location,int generation);

/* Compares two simulations to decide Pareto Dominance relation.*/
void Pareto_Comparison( struct siminfo *current_sim, 
        struct pset **ptr_to_pareto_member, struct pset **ptr_to_Frontier_first_location,
        int *on_list, int *stop_loop);

/* Functions to free dynamically allocated memory. */
void free_rest_list(struct psiminfo *start_simptr);

void free_rest_sim_list(struct siminfo *start_simptr);

void free_pareto(struct pset *start_pset);


/************** functions in PARETO_breed.c  ******************/

/* Selects parents from current generation of simulations and breeds 
offspring through either cross-over or mutation.*/
void breed(int gen, struct siminfo *current_gen_first_ptr, 
                double total_fitness, struct pset *Frontier_first_ptr);

/* Selects parents from current generation of simulations.*/
void par_selector(struct siminfo *current_gen_first_ptr, 
                double total_fitness, int select[POP_SIZE]);

/* Selects a parent's genetic operator.*/
void operator_selector( int cross[POP_SIZE], int mutate[POP_SIZE], 
                int *cross_cnt_ptr,int *mutate_cnt_ptr, int gen, int *mut_extra_ptr);

void parent_info(struct param parents[POP_SIZE], int select[POP_SIZE], 
struct siminfo *current_gen_first_ptr);

void crossover(int cross[POP_SIZE], struct param parents[POP_SIZE], int gen,
                int cross_cnt, int mut_extra, int *mut_yes_ptr);

void cross_operator(struct param *parent1, struct param *parent2, int gen);

void mutator(int mutate[POP_SIZE], struct param parents[POP_SIZE], int gen, 
        int mut_cnt, int mut_yes);

/* Checks to make sure each newly generated offspring parameterization
is unique and not already  on the Historical Pareto Frontier.*/ 
void new_individual_check(struct pset *Frontier_first_ptr, 
        struct param parents[POP_SIZE], int gen);
 
int frontier_check(struct pset *Frontier_first_ptr, struct param parent);

int current_pop_check(struct param parent_new, struct param parent_old);



/************* functions in PARETO_print.c ***************/

/* Reads in existing Pareto Frontier when pareto_evolve is called to 
start from a nonzero generation. See main.c for example call. */
void readin_pareto_frontier(char paretofilename[15], 
                                struct pset **ptr_to_Frontier_first_location);

void readin_assessment_vector(FILE *paretofileptr, struct pset *pareto_member,
                         int full_header);

void readin_psims( FILE *paretofileptr, struct pset *pareto_member);

int  AnotherMemberTest(FILE *paretofileptr);

/* Functions in main() or evaluation() that require printing characters.*/

void single_mutate(double *param_ptr,int j, int gen);

double mutation(int gen, int j, int rtype, double *param_ptr);

void initial_input(void);

void print_pareto(struct pset *Frontier_first_ptr, int gen);

/* Functions within breed that need print characters*/
int rinteger(double xvalue);

void offspring_output(struct param parents[POP_SIZE]);

double rand01(void);
   
/******************************************************/
/***************** EXTERNAL VARIABLES *****************/

        /* from breed() */

        int select[POP_SIZE];
        int cross[POP_SIZE];
        int mutate[POP_SIZE];
        struct param parents[POP_SIZE];

        /* from par_selector() */

        double cdf[POP_SIZE];

        /* from print_pareto() */

        char paretofile[100];

        /* from assign_fitness() */

        double max_niche_count[POP_SIZE];

        /* from main() */

        char critname[100];
        char paretofilename[100];

#endif
