/*******************************************************************
UserSettings.h
One of two header files that the user must edit before compiling
the Pareto_Evolve code (UserSettings.h, UserArrays.h).  
In this file the user defines the model's number of parameters and criteria,
the pathnames to the model's executable code, input file, and output file, and  
various optimization algorithm settings: type of criteria discrepancy measures
(binary or continous), whether to use an elitist search, search population
size, etc.
This file is 'included' in the general header file, ParetoEvolve.h.
********************************************************************        
    Editing History: 
    21 May 1998 - creation; J. Reynolds & M. Turley
    7 Sept 2005 - documentation revision, restructuring; R. Komuro 
    Feb 2006; Dec 2007 - documentation expansion, further restructuring to simplify use;
            Joel H. Reynolds, joel_reynolds@fws.gov
********************************************************************
NOTES:
The Pareto_evolve algorithm currently assumes that the criteria discrepancy
measures are either ALL binary or ALL continuous.  See other comments below.
*********************************************************************/

/*********************************************************************/
/****** MODEL DIMENSIONS ******/
/*********************************************************************/
#define NUMPARAMS 6      /* number of parameters in model being assessed */
#define NUMCRITERIA 12   /* number of criteria used in assessment */

/*********************************************************************/
/****** PATH AND FILENAMES for model executable and files ******/
/*********************************************************************/
    /* model executable.*/
#define SYST_CALL "c:\\Tree\\criteria\\Debug\\criteria"

    /* model input file (created by pareto_evolve) - ONLY EDIT PATH, NOT FILENAME! */
#define MODEL_INPUT_CALL "c:\\Tree\\input.txt"

    /* model output file (read by pareto_evolve)- ONLY EDIT PATH, NOT FILENAME! */
#define CRIT_OUT_CALL "c:\\Tree\\crit.out"

/*********************************************************************/
/****** FILENAMES for recording intermediate status of search.******/
/*********************************************************************/
    /* System call to create a permanent copy of the current model output file.
    The root filename in the 'destination' path will be appended with 
    the current generation number.  E.g., 'crit.out10*/
#define CRIT_COPY_CALL "copy c:\\Tree\\crit.out c:\\Tree\\crit"

    /* Path and root filename for storing intermediate Pareto Frontier Archives.
      The root filename will be appended with the current generation number. 
      ('Historical Pareto Frontier' = cumulative elitist archive of search.)
      #######
      WARNING: 
      The path must exist, ParetoEvolve cannot create it.
      I.e., create the necessary folders and subfolders where you want 
      ParetoEvolve to put the output files before running the code.*/                            
#define FRONTIER_CALL "C:\\TREE\\ParetoSets\\Paretoset"


/*********************************************************************/
/****** ASSESSMENT CONTROL SETTINGS ******/
/*********************************************************************/
#define BINARY 1   /* =1 if all discrepancy measures are binary, =0 if all are continuous */
#define GEN_TO_PRINT 500 /* how often to print out intermediate Pareto Frontier Archives
                            ('Historic Pareto Frontiers'). */

/*********************************************************************/
/****** OPTIMIZATION CONTROL SETTINGS ******/
/*********************************************************************/
    /***** SEARCH SIZE AND DURATION *****/
#define POP_SIZE 100   /* size of active search population */
#define GEN_NUM 1000   /* maximum number of generations to search */


/*********************************************************************/
    /***** FITNESS and GENETIC OPERATOR CONTROL SETTINGS *****/
/*********************************************************************/
   
    /* Maximum neighborhood radius in criteria space (for defining neighbors). */
#define MAX_SHARE_DIST 0.6   
/* If the assessment results of two different simulations are within this distance in the 
criteria space, then they will be considered 'neighbors' in calculating 
each one's niche_count. A simulation's fitness is adjusted by it's 
niche_count (reduced for more neighbors) in order to encourage the 
search algorithm to explore less visited regions. See share_fitness() 
in the file pomac_update.c.

If BINARY error measures are used, any MAX_SHARE_DIST < 1 will classify 
as neighbors only those simulations producing identical assessment vectors.

If CONTINUOUS error measures are used, the current code assumes that 
all criteria are measured on a scale of [0,1]. If this is not the case, 
criteria need to be rescaled. The value of MAX_SHARE_DIST should balance 
the maximum distance allowed when simulations differ on only a single 
criterion with the distance allowed when simulations differ on all 
criteria. 
For example, assuming continuous criteria defined on [0,1], the 
following values will define neighborhoods containing approximately 10% 
of the total criteria space hypervolume:
Number of Criteria      MAX_SHARE_DIST
        1                      .1
        2                      .18
        3                      .29
        4                      .60
        
NOTE: this measure has nothing to do with the calculation of Pareto 
dominance, only the fitness adjustment assigned a given simulation prior 
to selecting the next round of 'parents'.*/


#define WGT 0.5  /* WGT in (0,1]. Proportion by which to downweight the minimum fitness of the
                     i^th non-dominated rank in setting the baseline fitness of the 
                     (i+1)^th rank.  If WGT==1, then the baseline fitness of the (i+1)th
                     rank may equal the fitness of a i^th rank individual.
                     See share_fitness()in pomac_update.c 
                     NEEDS TO BE INVESTIGATED.*/
    
   /******** GENETIC OPERATOR SETTINGS ********/
#define MUTATE_PROB .6  /* Initial probability of selecting mutation rather 
                           than cross-over. This probability increases as
                           a function of the current generation and 
                           GEN_NUM. 
                           See operator_selector() in pareto_breed.c */

#define BETA 5   /* exponentional in mutation function; value taken from description of
                    nonuniform mutation in Michaelwicz, Z. 1996.  Genetic Algorithms + 
                    Data Structures = Evolutionary Programs (3rd) Springer-Verlag, p. 103.   */

/******************************************************************/





