/*************************************************************************
PARETO_breed.c

  This file contains the functions for selecting parents and breeding the 
  offspring for the next generation.                     

FUNCTIONS:   breed, par_selector, operator_selector, parent_selector, 
    crossover, cross_operator, mutator, single_mutate, mutation,
    offspring_output, ran1

*************************************************************************/

#include "ParetoEvolve.h" 

/**************************** breed ***************************************/
/* This function calls the different functions to breed the offspring, 
    the next generation's population of parameterizations. A parameterization 
    is a vector of parameter estimates.                                   */
/**************************************************************************/

void breed(int gen, struct siminfo *current_gen_first_ptr, 
        double total_fitness, struct pset *Frontier_first_ptr)
{
    extern int select[POP_SIZE];    /* array of parents selected to breed. */
    extern int cross[POP_SIZE];     /* array of parents to receive a crossover 
                                        out of those selected to breed. */
    extern int mutate[POP_SIZE];    /* array of parents to receive a mutation 
                                        out of those selected to breed. */
    int cross_cnt=0;        /* count of number of parents to be crossed. */
    int mutate_cnt=0;       /* count of number of parents to be mutated. */
    int mut_extra=0;        /* extra parent if crossover count is odd. */
    extern struct param parents[POP_SIZE];  /* list of parents with their 
                                                parameterization information */
    int mut_yes=0;          /* flag that indicates a mutate selection is used 
                                for a crossover when crossover count is odd. */
 
 

    /* Randomly select the parents from the current generation's population that will 
        breed the next generation. */

        par_selector(current_gen_first_ptr,total_fitness,select);

    /* Copy the parameterization information of the selected parents from 
        the simulation list into a parents' list */

        parent_info(parents,select,current_gen_first_ptr);

    /* Randomly select a subset of the selected parents which will receive a crossover 
        and those which will receive a mutation. */

        operator_selector(cross,mutate,&cross_cnt,&mutate_cnt,gen,&mut_extra);

    /* Each parent selected to crossover is crossed with another parent for randomly 
        selected parameter estimates of their parameterizations to create two new 
        parameterizations for each pair of parents crossed.   */

        crossover(cross,parents,gen,cross_cnt,mut_extra,&mut_yes);

    /* Each parent selected to mutate is mutated for randomly selected parameter 
        estimates of their parameterization to create a new parameterization. */

        mutator(mutate,parents,gen,mutate_cnt,mut_yes);

    /*  Check that the new pararemetizations are unique. That is, check that they 
        are not already in the Historical Pareto Frontier or as duplicates in the 
        new set of parameterizations. */

        new_individual_check(Frontier_first_ptr,parents,gen);

    /* Print the new set of parameterizations to a model input file to be run as 
        the next generation's population (located in PARETO_print.c). */

        offspring_output(parents);

}   /* end of breed  */




/*************************  par_selector  **********************************/
/* This function selects with replacement from the current population the 
    parents that will breed the offspring. The parents are a population of 
    size = POP_SIZE. An individual's chance of selection is based on the 
    probability equal to their fitness/total fitness of population. A cumulative 
    distribution function is created using the simulations' fitness measures.
    Random numbers from a uniform[0,1] distribution are selected. Each random 
    number is evaluated to see in which simulations' c.d.f. range the number 
    landed. When the random number occurs in a simulation's range, that 
    simulation is selected to be a parent. The function returns a vector of 
    the selected parents (some individuals may be chosen more than once) back 
    to breed to be used in crossover and mutate functions. */
/***************************************************************************/

void par_selector(struct siminfo *current_gen_first_ptr, double
         total_fitness, int select[POP_SIZE])
{
    int i,j;                        /* loop control variables */
    struct siminfo *current_sim;    /* pointer to current simulation */
    extern double cdf[POP_SIZE];    /* cumulative distribution function */
    double randpick;                /* random, uniform [0,1] number to select parent */
    int select_flag;                /* indicator of whether a parent is selected */




    /* Initialize the cumulative distribution function and the vector of the selected 
    parents. */

        for (i =0; i<POP_SIZE;i++)
        {
         cdf[i]=0.0;
         select[i]=0;
        }
    
    /* Start at the top of the current generation. Calculate the first simulation's 
    proportional fitness. Increment to the next simulation. */

        current_sim = current_gen_first_ptr;
        
        cdf[0] = current_sim->fitness/total_fitness;
        current_sim = current_sim->next;

    /* Loop through the rest of the current generation's parameterizations, calculating its 
    cumulative distribution function based on each parameterization's proportional fitness. */

        for (i=1; i < POP_SIZE; i++)
        {
          cdf[i] = current_sim->fitness/total_fitness + cdf[i-1];
          current_sim = current_sim->next;
        }



    /* Select n individuals with replacement from the current generation to be parets to breed 
    the next generation.*/

        for (i=0; i < POP_SIZE; i++)
        {
     /* Initialize flag that terminates loop when a parent is selected. */
    
          select_flag = 0;

     /* Pick a random number from a uniform [0,1] distribution. */

          randpick = rand01();
    
     /* Check to see if the random number occurs in the first parameterization's
        cdf range. If yes, the first parameterization is selected as a parent. */

          if( (randpick <= cdf[0]) && (randpick >= 0) )  select[i] = 0;

     /* Otherwise, loop through the rest of the parameterizations to see where the 
        random number occurs in the parameterizations' cdf ranges. When the random number 
        occurs in a parameterization's range, that parameterization is selected as a parent. */

          else
            {
              j=1;
              while (j != POP_SIZE)
                {
                  if( (randpick <= cdf[j])&&(randpick > cdf[j-1]))
                  {
                    select[i] = j;
                    break;
                  }
                  j++;  
                }       /* end of while j loop */
            }       /* end of else loop */ 

        } /* end of i loop */


    /* Return to the calling function, breed. */


}   /* end of par_selector  */




/************************* parent_info **********************************/
/* This function copies the parameterization information of the selected 
parents from the current generation's parameterization list 
(lexicographically ordered list) into a parents' list.                  */
/************************************************************************/

void parent_info(struct param parents[POP_SIZE], int select[POP_SIZE], 
        struct siminfo *current_gen_first_ptr)
{
    int i,j,k;                      /* loop control variables */
    struct siminfo *current_sim;    /* pointer to current simulation (parameterization) */



    /* Loop thorugh the current generation. */


    for(i=0; i < POP_SIZE; i++)
     {

    /* Start at the top of the current simulations' lexicographically ordered list. */

       current_sim = current_gen_first_ptr;
       j=0;

    /* Match each selected parent with its location in the ordered list by looping 
       through list. The j counts the location in the list. */

       while(j++ < select[i])
         {
            current_sim = current_sim->next;
         }

    /* When the simulation location matches the selected parent, copy that simulation's 
       parameterization into the parents' list.*/

       for(k=0; k < NUMPARAMS; k++)
         {
            parents[i].par[k]=current_sim->paramvalues.par[k];
         }
     }

    /* Return to the calling function, breed. */

}   /* end of parent_info  */



/************************* operator_selector ******************************/
/* This function randomly chooses which parents will receive a crossover 
    and which will receive a mutuation. As the generations increase, the 
    number of crossovers decreases and subsequently the number of mutations 
    increases. select[i] equals the number of an individual in the current
    population. This function returns two arrays: 1) the selected parents to 
    receive crossovers, 2) the selected parents to receive mutations. These 
    arrays have lengths = POP_SIZE available, but they will not necessarily 
    be that length since the parameterizations are divided randomly between 
    the two arrays. This function also returns the counts for crossovers 
    and mutations and the extra mutation to be used with a crossover.     */
/**************************************************************************/

void operator_selector( int cross[POP_SIZE],int mutate[POP_SIZE], int *cross_cnt_ptr,
                       int *mutate_cnt_ptr, int gen, int *mut_extra_ptr)
{

    int i;              /* loop control variable */
    double crosscut;    /* probability that a selected parent receives a crossover */


    /* Assign the probability that a selected parent receives a crossover. The crossover 
    probability starts at 0.67 and decreases to 0,0.0 over the number of generations.
    (FLAG!!! We may want to change later). */   
 
       crosscut=( (double)GEN_NUM - (double)gen)/(1.5*(double)GEN_NUM);

    /* Loop through each selected parent and decide whether it will receive a crossover 
       or a mutation. i indexes the population of selected parents.  */ 

    for (i=0; i < POP_SIZE; i++)
     {
      if( rand01() <= crosscut )
        {

    /* Randomly select a number from a uniform[0,1] distribution and check that it is less 
    than or equal to the crossover probability. If yes, then assign that selected parent 
    to receive a crossover. i indicates a parameterization in selected parents' list. */

          cross[*cross_cnt_ptr] = i;

    /* Count the number of parents assigned to receive a crossover. */

          (*cross_cnt_ptr)++;
        }

    /* Otherwise, assign that selected parent to receive a mutation. */

      else
        {
          mutate[*mutate_cnt_ptr] = i;

    /* Count the number of parents assigned to receive a mutation. */

          (*mutate_cnt_ptr)++;
        }

    }  /* end of i loop */


    /* If there are an odd number of parents to receive crossovers, send a mutate 
    parent to the crossover list. */

    if(*cross_cnt_ptr%2 != 0)    *mut_extra_ptr = mutate[0]; 
 

    /* Return to the calling function, breed. */
 

}   /*  end of operator_selector  */



/*************************  crossover  *********************************/
/* This function cycles through the parents' list, applying crossovers 
to those parameterizations designated as such by operator_selector.  */
/************************************************************************/

void crossover(int cross[POP_SIZE], struct param parents[POP_SIZE], int gen,
        int cross_cnt, int mut_extra, int *mut_yes_ptr)
{
    int i;                  /* loop control variable */
    int temp_cross_cnt;     /* local variable of crossover count */



    /* Assign crossover count to local variable. */
 
        temp_cross_cnt=cross_cnt;
    

    /* Establish whether there are an even number of pairs of parents to crossover.
    It takes two to "reproduce" by crossover.  */
 

    if( (cross_cnt % 2) != 0)  
     {

    /* If there are an odd number of individuals, move a parent from the mutation 
    list to the crossover list. */

        cross[cross_cnt] = mut_extra;

    /* Increment the local crossover count. */

        temp_cross_cnt++;

    /* Send a flag to indicate that the additional mutation parent has been used in the
    crossover process. */

        *mut_yes_ptr = 1;   
     }

   /* Match up the parents into pairs then cross over some of their parameter estimates.  */

     for(i=0; i < temp_cross_cnt; i +=2)
      {

        cross_operator( &parents[cross[i]], &parents[cross[i+1]],gen);
 
      } /* end of crossover count loop */

    /* Return crossed parents and mut_yes flag to the calling function, breed. */ 

}   /* end of crossover  */




/************************ cross_operator  *******************************/
/* This function does the cross over between two parents.               */
/************************************************************************/

void cross_operator(struct param *parent1, struct param *parent2, int gen)
{

    int j;              /* loop control variable */
    struct param temp;  /* temporary variable in which to store a parent */
    int crossed = 0;    /* indicator that a parameter estimate has been crossed 
                            between two parents */
    int same_parent;    /* counter of the number of equal parameter estimates between 
                            two parents */

    
    /* Check that parents are not the same parameterizations.  */

        same_parent = 0;
    
    /* Count how often the two parents' have equal parameter estimates. */
    
        for(j=0; j< NUMPARAMS; j++)
        {
          if (parent1->par[j] == parent2->par[j]) same_parent++;
        }

    /* If the equality counts equal the number of parameters, they are the same
    parent, so mutate one parent. */

        if(same_parent == NUMPARAMS) 
        { 
          for(j=0; j< NUMPARAMS; j++)
            {
             single_mutate(&parent2->par[j],j,gen);
            }
        }

    /* Begin crossover of two parents. Loop until at least one parameter 
    estimate is crossed. */

    while(crossed == 0)
    {
      for(j=0; j < NUMPARAMS; j++)
      {

    /* Randomly select a number from a uniform[0,1] distribution of whether to
    cross a parameter estimate. */

        if ( rand01() <= 0.5)
         {
            temp.par[j] = parent2->par[j];
            parent2->par[j]=parent1->par[j];
            parent1->par[j]=temp.par[j];

    /* Count the number of parameter estimates crossed.*/

            crossed++;
         }  /* end of if loop */

    /* Do not allow all parameter estimates to be crossed, otherwise the same 
    two parents will result. */

        if(crossed==3) break;

      } /* end of j loop */
    }   /* end of while loop */
 
    /* Return crossed parents to calling function, crossover. */

}   /* end of cross_operator  */



/************************** mutator *************************************/
/* This function cycles through the parents' list, applying mutate to
   those individuals designated as such by the genetic_selector.        */
/************************************************************************/

void mutator(int mutate[POP_SIZE], struct param parents[POP_SIZE], int gen, 
    int mut_cnt, int mut_yes)
{
    int i,j;                        /* loop control variable */
    int mutated;                    /* count of number of mutated parameter estimates  */
    struct param current_parent;    /* local variable of parent selected to be mutated */
    

    /* Start from top of parents to be be mutated list (mut_yes = 0). If a parent to be 
    mutated was used with the crossover function (mut_yes = 1), start from the second 
    parent in the parents to be mutated list. */
  
       for( i = mut_yes; i < mut_cnt; i++)
        {
  
    /* pass the parameterization information of current parent to be mutated to a local 
    variable */
    
           current_parent = parents[mutate[i]]; 

    /* Initialize the count of how many of the current parent's parameter estimates are 
    mutated. */

        mutated = 0;

    /* Begin the mutation of the current parent. Repeat loop until at least one mutation 
    occurs in the cycle of mutating possibly all of the parameter estimates. */

        while(mutated == 0) 
            {

    /* Loop through all of the parameter estimates in a parameterization. */

            for(j=0; j < NUMPARAMS ; j++)
                {

    /* Randomly select from a uniform [0,1] distribution of whether to mutate a parameter 
    estimate.  */

                if (rand01() <= 0.50)       
                    {   
                        single_mutate(&current_parent.par[j],j,gen); 
                        
    /* Count that at least one parameter estimate has been mutated in the current parent. */ 

                        mutated = 1;
                    }

                }    /* end j parameters loop   */

    /* Exit while loop and examine next parent when at least one parameter estimate in the 
    current parent was mutated. */

            }   /* end of while mutated loop  */

    /* Exit for loop when all of the parents to be mutated have been mutated. */
 
        }  /*  end i loop   */

    /* Return mutated parents to the calling function, breed. */

}   /* end of mutator  */


/************************ new_individual_check  *************************/
/*  This function checks that the new parameterizations bred for the next 
generation are not already in the Historical Pareto Frontier or in the 
new parameterizations list. This function makes sure that all new 
parameterizations are unique so that none are checked more than once.   */
/************************************************************************/

void new_individual_check(struct pset *Frontier_first_ptr, 
    struct param parents[POP_SIZE], int gen)
{
    int i,j,k;      /* loop control variables */
    int unique;     /* indicator of whether parameterization is on a list or not */



    /* Loop through the new parameterizations list, checking each parameterization 
    against the Frontier and the new paramterizations list. */

    i=0;

    while(i != POP_SIZE )
      {

    /* First check a parameterization against the Frontier.  Initialize the indicator of 
    the paramterization's uniqueness against the Frontier list. */

        unique=0;

        while(unique == 0)
        {

    /* Compare a parameterization against the Frontier. The function returns a 1 if the
    parameterization is unique and it returns a 0 if the parameterization is in the 
    Frontier list already. */

            unique = frontier_check(Frontier_first_ptr,parents[i]);

    /* If a parameterization is already on the list, it is mutated and the while loop 
    is repeated. */

            if(unique == 0)
            {
              for(k=0; k < NUMPARAMS; k++)
                {
                  single_mutate(&parents[i].par[k],k,gen);
                }
              break;
            }  /* end of if (unique == 0) loop */

 
    /* If the parameterization is unique to Frontier, check that it does not already occur 
    in the new paramterizations list. */

            else
            {
    /* If the parameterization is the first in the list, it is unique for this list. */

              if(i == 0)
                {
                  unique = 1;
                  break;
                }
        /* Otherwise, check the rest against the list. */
              else
                {
                  for(j= 0; j < i; j++)
                    {

                      unique = current_pop_check(parents[i],parents[j]);

    /* If a parameterization is already on the list, it is mutated and the while loop 
    is repeated. */
                        if(unique == 0)
                            {
                              for(k=0; k < NUMPARAMS; k++)
                                {
                                  single_mutate(&parents[i].par[k],k,gen);
                                }
                              break;
                            }   /* end of if (unique == 0) loop */

                    }   /* end of for j loop */
                 
                }     /* end of else rest loop */

            }   /* end of else check new parameterizations list loop */

        }   /* end of while unique loop */
        i++;

    } /* end of while POP_SIZE loop */

    /* Return checked new parameterizations to calling function, breed. */

}   /* end of new individual_check */





/**************************frontier_check ***************************/
/* This function compares a single new parameterization against the 
    Historical Pareto Frontier to make sure the parameterization is 
    not already in the Frontier.                                    */
/********************************************************************/

int frontier_check(struct pset *Frontier_first_ptr, struct param parent)
{

    int i;
    int equal_cnt=0;
    struct pset *pareto_ptr;
    struct psiminfo *psim_ptr;
 


    /* Start at the top of the Historical Pareto Frontier list. */

        pareto_ptr = Frontier_first_ptr;

    /* Loop through and compare to all of the Frontier list until a 
    match of the parameterization is found in the list or the end of 
    the list is reached. */

        while(pareto_ptr!=NULL)
        {
    /* Copy the Frontier member's parameterization into a local variable to compare. */

         psim_ptr = pareto_ptr->psim;

    /* Loop through all of the parameterizations associated with a Frontier member. */

         while(psim_ptr!=NULL)
          { 
    /* Initialize the indicator of uniqueness to Frontier list. */

           equal_cnt = 0;

    /* Loop through all of the parameter estimates per parameterization. If the 
    parameter estimates are equal, give the equality count a point.*/

           for(i=0; i < NUMPARAMS; i++)
            {
             if(parent.par[i] == psim_ptr->paramvalues.par[i] )  equal_cnt++;
            }

    /* Check to see if the current parameterization and Frontier member parameterization
    are equal. If they are not equal, check the current against next parameterization 
    for this Frontier member. If they are equal, return a nonuniqueness flag (function =0) 
    to the calling function, new_individual_check. */

           if(equal_cnt != NUMPARAMS)
            {
             psim_ptr = psim_ptr->next;
            }        
           else 
            {   
              return (0);
            }

      } /* end of while psim_ptr loop */

    /* If not the current paramterization is not equal to any parametrizations in the first 
    Frontier member, investigate the paramterizations in the next Frontier member. */

    pareto_ptr = pareto_ptr->next;

    }   /* end of while pareto_ptr loop */
 
    /* Else, the parameterization does not already exist in the Frontier list. Return an 
    uniqueness flag (function = 1) to the calling function, new_individual_check. */ 

    return (1); 
 
}   /* End of frontier_check. */



/********************** current_pop_check **************************=**/
/*  This function compares a single new parameterization against the 
    new parameterizations list to make sure the new parameterization 
    is not already on the list.                                     */
/********************************************************************/
int current_pop_check(struct param parent_new, struct param parent_old)
{
    int k;
    int equal_cnt;
 

    /* Initialize the indicator of uniqueness to the new parameterizations list. */

        equal_cnt = 0;

    /* Compare the new parameterization with one from the list, looping through the 
    parameter estimates. Count the occurrences of equal parameter estimates. */

        for(k=0; k < NUMPARAMS; k++)
         {
            if(parent_new.par[k]== parent_old.par[k]) equal_cnt++;
         }

    /* If the paramterizations are equal, this parameterization is already on the new 
    parameterizations list. Return a nonuniqueness flag (function = 0) to the calling 
    function, new_individual_check. */

    if(equal_cnt == NUMPARAMS)
      {
        return (0);
      }

    /* If the new parameterization is different, it may be added to the new 
    parameterizations list. Return an uniqueness flag (function = 1) to the calling 
    function, new_individual_check. */

    else
      {
        return (1);
      }


}   /* End of current_pop_check. */



  

/*********************************** rand01 **********************************/
/* This function is a random-number generator for numbers from a uniform[0,1]
    distribution. It uses the C function rand() which produces numbers 
    from a range of (0,RAND_MAX). RAND_MAX is defined somewhere in the system.
    This function is used in several functions of this code.                 */
/*****************************************************************************/
    
double rand01(void)
{
    static int firstcall=1;

    /* Seed the pick for rand() by the time clock, so each pick is unique. */

    if(firstcall)
    {
      srand( (unsigned int) time (NULL) );
      firstcall = 0;
    }
    
    /*  Pick a random number, find its two digit remainder and make that 
    number into a number between (inclusive) zero and one. Return the fraction 
    to the calling function. */

    return (rand() % 101)/100.0 ;
 
}       /* End of rand01 */
