/**********************************************************************
        PARETO_main.c            last edited 30 Sept 98 by JHR, MCT
                                     6 Oct 08 JHR remove debug flags
**********************************************************************
USAGE:  
1. Prior to using pareto_evolve, the user needs to 
        a. define error measure types and array limits in 
        UserSettings.h.
        b. define parameter names and search ranges, criteria names and 
        assessment measure information in UserArrays.h.
        c. set population size (POP_SIZE) and number of generations 
        (GEN_NUM) to allow solution search to evolve, model input and 
        output filenames and pathways, and system call to execute model 
        in ParetoEvolve.h.
        d. Compile code. The code has currently been tested for 
        compilation and execution using MS Visual C/C++ 6.0, 
        Professional Edition. 
        NOTE on compiling: be sure to compile for release, not 
        debugging, in order to improve efficiency of code.
2. Pareto_evolve can be run in three different modes:
        a. CONTINUOUS Search: to run the search for all GEN_NUM 
        generations from one call, simply (i) open the windows explorer to 
        the directory containing the compiled executable 
        (pareto_evolve.exe) and double click , (ii) open the Start 
        menu, select the Run command, and either enter the pathway to 
        the executable or browse and select it, or (iii) open a DOS 
        window, move to the directory containing the executable, and 
        type (w/o quotations) 'pareto_evolve' followed by the ENTER key.
         
        b. FINISHING an INCOMPLETE Search: If pareto_evolve was 
        stopped by the user prior to completing all GEN_NUM generations, 
        you can finish the search by restarting at the last generation 
        for which the Pareto Frontier (revealed up to that point of the 
        search) was printed out to file. You also need to have the 
        latest file of model outputs. 
        E.g., if the last generation for which the Pareto Frontier 
        was printed out was 350 (so there is a file pareto_set.350 
        in the directory set in FRONTIER_CALL in the ParetoEvolve.h and, 
        presumably, the model output file from the 351st generation of 
        simulations in the directory set in CRIT_OUT_CALL), 
        open a DOS window, move to the directory containing the 
        executable, and type (w/o quotations) 'pareto_evolve 351' 
        followed by the ENTER key. 
        This tells pareto_evolve to restart from the 351st generation by 
        locating and reading in pareto_set.350  and the (presumably 
        latest) model output file. It will then update the Pareto 
        Frontier, breed a new generation of simulations, etc., and 
        continue the search until the GEN_NUMth generation.
        
        c. DOING a PARTIAL Search: to run the search for only a 
        proportion of the complete GEN_NUM generations, say the first 500 
        generations when GEN_NUM is 2000, open a DOS window, move to the 
        directory containing the executable, and type (w/o quotations) 
        'pareto_evolve 0 500' followed by the ENTER key.
        This tells pareto_evolve to start at generation 0 and stop after 
        the 500th generation. You can later complete the search by 
        executing 'pareto_evolve 500'.
        
        SUMMARY: From the DOS command line, the general form of calling 
        pareto_evolve is
                                pareto_evolve startgen stopgen
        If only one argument is given, pareto_evolve reads it as startgen 
        and sets stopgen=GEN_NUM. If no arguments are given, 
        pareto_evolve sets startgen=0 and stopgen=GEN_NUM. 
        
        NOTE: GEN_NUM is set prior to compilation (in ParetoEvolve.h) 
        and influences both how the genetic operator selection probabilities 
        change as the generations progress and the how the non-uniform 
        mutation probability distribution gets rescaled as the 
        generations progress. 
***********************************************************************
Output Produced:
PRIMARY RESULTS:
Every GEN_TO_PRINT (set by the user in UserSettings.h), pareto_evolve 
creates a file containing the Pareto Frontier as so far revealed 
(default name is 'pareto_set.GEN#') and the current generation's model 
output file (default name is 'crit_out.GEN#') that one can later 
investigate the evolution of the Pareto Frontier as well as of the 
parameter space search.
TEMPORARY RESULTS:
Every generation pareto_evolve creates a new model input text file, and 
the model call creates a new output text file. These are  
overwritten during the next generation.
**********************************************************************
General Outline of Algorithm
#################################################################
 pareto_evolve generates a model's Pareto Frontier (aka Pareto Optimal Set) 
 by using an Evolutionary Computation Optimization algorithm to search 
 the model's feasible output space (as defined by the selected set of 
 criteria and the parameter search ranges). The basic algorithm follows 
 the Non-Dominated Ranking algorithm of Srinivas & Deb (Srinivas, N., 
 and K. Deb, 1995. Multiobjective Optimization Using Nondominated 
 Sorting in Genetic Algorithms. Evolutionary Computation 2(3): 221-248.), 
 though with the inclusion of a fitness sharing/nich_count method (see 
 Srinivas & Deb, and references therein) to maintain search diversity in the 
 objective space.
 
 pareto_evolve can be run with either ALL binary criteria error measures 
 or ALL continuous criteria error measures.
 
 *******************************************************************
        Outline:
        in evaluation()
        Reads in the current crit.out file
        Assesses each parameterization's criteria results
        Assigns each parameterization its NonDominated Rank relative 
        to the rest of the current generation's parameterizations.
        Assigns baseline fitness to each parameterization based on its 
        NonDominated Rank, then adjusts that fitness based on its 
        niche_count in the objective space.
        Updates the 'Historical' Pareto Frontier as necessary using 
        NonDominated Rank 1 parameterizations from current generation.
        
        in breed()
        Selects parents for next generation based on fitness.
        Assigns genetic operator to each parent
        Breeds next generation (offspring)
        Writes offspring to model input file
        
        back in main()
        If appropriate, write out Pareto Frontier and save criteria 
        output results from this generation.
        Check that Search should continue (generation==GEN_NUM or Pareto 
        Frontier contains element simultaneously satisfying all 
        criteria).
        If search should continue, call model to run next generation's
        simulations.
        
        Loop back to evaluation() to repeat search loop.
***********************************************************************/

#include "ParetoEvolve.h"


int main(int argc, char **argv)

   /* argc - number of arguments passed, including program call*/
   /*  **argv - ptr to list of strings */
{

 /*================      DECLARATION      =================*/

        FILE *ip;                                                               /* input file pointer */

        struct pset *Frontier_first_ptr=NULL;                   /* ptr to start of Historical Pareto 
                                                                                                                Frontier list*/

        struct siminfo *current_gen_first_ptr=NULL;             /* ptr to start of current generation's 
                                                                                                                simulation results list*/

        int generation;         /* current generation number */
        int stop_gen;           /* flag to stop the search at a generation number less than GEN_NUM */
        int pareto_flag;        /* flag to stop search when one or many parameterizations 
                                                        satisfy all of the criteria */

        double total_fitness;   /* cumulative fitness of entire generation  */
        extern char critname[100]; /* buffer for renaming crit.out file 
                                        & pathway*/
        extern char paretofilename[100];/* buffer for saving pareto.set file 
                                                & pathway.*/ 
 
        

        /* Check for !0 Generation Starting point and partial stopping point */

  stop_gen = GEN_NUM;
  generation=0;

   if (argc > 1) 
    {
        /* Read generation to start at other than 0. The generation to restart is 
        the second argument in the command line.  */

                char *a1 = argv[1];  
                sscanf(a1, "%d", &generation);

        /* Suppose the search is stopped after generation 20. To restart, execute from 
        generation 21. That is, paretorun.exe 21 and the search will read in the file, 
        pareto.set20, for the latest updated Historical Pareto Frontier. Restart must 
        be after generation 0.  */

                sprintf(paretofilename, "%s%d",FRONTIER_CALL,generation-1);
                readin_pareto_frontier(paretofilename, &Frontier_first_ptr);

        /* A user can also set the generation to stop the search. It must be a number 
        less than GEN_NUM. For example, to run the program for generation 101 to 300,
        execute by a command such as: paretorun.exe 101 300. The stopping number, here 
        it = 300, is the third argument in the command line. */

      if (argc == 3) 
      {
                char *a2 = argv[2];
                
        /* Read generation to stop at */

                sscanf(a2, "%d", &stop_gen);
                if(stop_gen > GEN_NUM) 
                        { 
                                printf(" To run search to %d generations,",stop_gen);
                                printf(" increase GEN_NUM in evolve.h and recompile.\n");
                                exit(1);
                        }
      }         /* End of scanning command line for stop. */

   }    /* End of scanning command line for restart. */



        /* Initialize stopping flags */

                pareto_flag=0;


        /* If the search is at genration 0, generate a population of parametrizations 
                (initial input values) to run with the model.  */

                if (generation == 0) initial_input();
 

        /* Call model to do a round of simulations to initialize crit.out. If an error 
                occurs while running the model, send a message and terminate the search. */

                if (system(SYST_CALL)==-1) 
                {
                        printf(" error during model runs");
                        exit(1);        
                } 



                /* The evolutionary computation search loops through the number of generations 
                to stop_gen or until the pareto flag signals that the search has been satisfied. 
                Each generation is evaluated for its population's ability to satisfy the criteria. 
                The results are printed to a file. The current generation breeds offspring to be 
                the next generation. The search uses the offspring in the model to produce 
                criteria results for the offspring and the while loop repeats. */

  while ((generation <= stop_gen) && (pareto_flag!=1)) 
   {     
   
        /* Continue the evolutionary computation search. Open the file with the model's 
        criteria results for the current generation. The file includes both the 
        paramterizations and the criteria vectors for the current population. */

                ip = fopen(CRIT_OUT_CALL,"r");

        /* Initialize total fitness for each generation. */

                total_fitness= 0.0;

        /* Evaluate the current generation for its ability to satisfy the criteria by Pareto 
         optimization.  */ 

                evaluation(&Frontier_first_ptr,&current_gen_first_ptr, ip, &pareto_flag,
                        &total_fitness,generation);

        /* Close the crit.out file.  */

     if(fclose(ip)!=0) printf("In main, can not close crit.out file.\n");

        /* Copy the updated Historical Pareto Frontier to a file for the current generation.
         The first if statement sets how often this file is printed. E.g., if GEN_TO_PRINT =10,
         every tenth generation is printed. */ 

   if( (generation % GEN_TO_PRINT) == 0)
        {
      print_pareto(Frontier_first_ptr,generation); 
      sprintf(critname,"%s%d",CRIT_COPY_CALL,generation);
      system(critname); 
        }

   /* If the search for one or many paramterizations that satisfy the criteria has not been 
   satisfied, continue the optimization by breeding a new generation of parameterizations to
   be used with the model. */

   if (pareto_flag != 1)   
   {
 
        /* Generate the offspring. This function also makes sure that the new parameterizations have 
           not already been included in the Historical Pareto Frontier. */

     breed(generation,current_gen_first_ptr,total_fitness,Frontier_first_ptr);
 

        /* Call the model to do the next round of simulations with the new parameterizations. 
        If an error occurs while running the model, send a message and terminate the search.*/
 
        if (system(SYST_CALL)==-1) 
                {
                        printf(" error during model runs");
                        exit(1);
                }
 

    } /* End of breeding new generation and calling next round of simulations with the model.*/

        /* Increment generation counter for the next generation and return to top of while loop. */

                generation++;
 
   } /* End of multiple generations loop. */

        /* Free up memory used to store the Historical Pareto Frontier. */

     free_rest_sim_list(current_gen_first_ptr);
     free_pareto(Frontier_first_ptr);

        /* Signal the end of the evolutionary computation search with a bell! */

         printf("\007");

} /* End of main. */


 
