/**********************************************************************
  PARETO_update.c

    
   Collection of functions, called by pareto_evolve.c, to read in current 
   crit.out, calculate each simulation's fitness, and update the Historical 
   Pareto Frontier. 

  
FUNCTIONS: evaluation, read_input, criteria_assess, ordered_add_to_current_gen,
    insert, tie_test, compare, assign_fitness, criteria_distance,
    sharing, non_dominated_ranking, crit_comp, share_fitness,
    Pareto_Frontier_update, Pareto_Comparison, free_pareto, 
    free_rest_list, print_pareto
*********************************************************************
original code: fall 94 - 4/95 JHR. last revised: 6 July 98 mct
                                     6 Oct 08 JHR remove debug flags
*********************************************************************/
#include "ParetoEvolve.h"


/************************  evaluation  ******************************/
/* Describe stages of this function, explain process.             
  

    struct siminfo **ptr_to_head_location in the function declaration 
    variable is the address of ptr to current_gen_first. Siminfo structure 
    stores information about parameterizations, criteria vectors, 
    assessment vectors, nondominated rankings, niche counts, and fitnes
    measures.                                                        */
/*********************************************************************/

void evaluation(struct pset **ptr_to_Frontier_first_location, 
        struct siminfo **ptr_to_head_location, FILE *input_ptr, 
        int *pareto_flag_ptr, double *total_fitness_ptr, int generation)
{
  struct siminfo *initial_sim_ptr; /*structure of initial simulation in current population*/
  struct siminfo *latest_sim_ptr;  /*structure of simulation currently being processed*/


  int erc=0;    /* error flag for input reading function */
  int i;        /* generic counter */
  int pcount=0; /* place counter of number of simulations that have been read in */ 
  int temp;     /* variable to print nondominated rank */


  /* Initialize the structure and allocate memory for initial simulation */

    initial_sim_ptr=(struct siminfo *)malloc(sizeof(struct siminfo));
    if(initial_sim_ptr == NULL) printf("out of stack memory\n");
   
 
    /* Throw out headers in simulation results file. */

    for (i=0; i < (NUMPARAMS+NUMCRITERIA); i++) fscanf(input_ptr, "%*s");

    initial_sim_ptr->NDrank=0;
    initial_sim_ptr->niche_count=1.0;
    initial_sim_ptr->fitness=0.0;
    initial_sim_ptr->next=NULL;

    /* Read in the parameterization and criteria results for the initial 
    simulation (individual) in the current generation (population). */
 
    read_input(&erc, input_ptr,initial_sim_ptr,&pcount);
    if(erc == 1)
      { 
     printf("In evaluation, error reading input!\n");
     exit(1);  
      }                  

    /* Assess the initial simulation in terms of the criteria.  */

    criteria_assess(initial_sim_ptr);    

    /* Assign a pointer to the top of the list of the current 
    population. */

    *ptr_to_head_location=initial_sim_ptr;
 
    
    /* Loop through the rest of the population. */

    while(pcount != POP_SIZE)
     {
   
    /* Allocate memory. */

        latest_sim_ptr=(struct siminfo *)malloc(sizeof(struct siminfo));
      if(latest_sim_ptr == NULL) printf("out of stack memory");


    /* Initialize the elements in the structure. */

      latest_sim_ptr->NDrank=0;
      latest_sim_ptr->niche_count=1.0;
      latest_sim_ptr->fitness=0.0;
      latest_sim_ptr->next=NULL;

     /* Read in the parameterization and criteria results for the latest 
     simulation (individual) in the current generation (population). */

      read_input(&erc, input_ptr,latest_sim_ptr,&pcount);
      if(erc == 1)
        { 
      printf("In evaluation, error reading input!\n");
      exit(1);  
        }

     /* Assess the latest simulation. */
      
      criteria_assess(latest_sim_ptr);  
      
     /* Add the latest simulation to a lexicographically ordered list 
      of the current genration's simulations that have already been 
      processed. The list will be used for nondominated ranking. */
      
     ordered_add_to_current_gen(latest_sim_ptr, ptr_to_head_location);
      
    }   /* End of while pcount loop. */


    /* Once ordering is completed, assign fitness measures to 
    individuals in terms of nondominated ranking and niche count. */

    assign_fitness(*ptr_to_head_location, total_fitness_ptr);
    temp=(**ptr_to_head_location).NDrank;

    /* Update the Historical Pareto Frontier based on the current 
    population's results. In the Pareto_Frontier_update variable 
    list:   *ptr_to_head_location in the function declaration 
    variable is the address of ptr to current_gen_first.    */

    Pareto_Frontier_update(*ptr_to_head_location,
         ptr_to_Frontier_first_location, generation); 

}   /* End of evaluation, returns to main. */
    


/******************  read_input *****************************************/
/*  This function reads in a simulation's parameterization used with the 
model to produce the criteria results and criteria results. A paramterization 
is a vector of model parameter estimates.                               */
/************************************************************************/

void read_input(int *erc_ptr,FILE *input_ptr, 
        struct siminfo *latest_sim_ptr,int *pop_ptr)
{
   int i;
   int scanpar,scancrit;

   /* Read in the parameterization for the simulation. */

   for (i=0; i<NUMPARAMS; i++) 
     {
      scanpar=fscanf(input_ptr, "%lf", &latest_sim_ptr->paramvalues.par[i]);
      if(scanpar != 1) *erc_ptr=1;
     }


    /* Read in the criteria results for the simulation. */

   for (i=0; i< NUMCRITERIA; i++) 
     {
      scancrit=fscanf(input_ptr, "%lf", &latest_sim_ptr->cresults.crit[i]);
      if(scancrit != 1) *erc_ptr=1;
      /* if (DEBUGFLAG==1) printf("crit =%lf ",latest_sim_ptr->cresults.crit[i]);*/
     }
 
   /* Increment the generation (population) pointer to the next individual 
   and pass back to the calling function. */
   
   *pop_ptr= *pop_ptr + 1;
   
}   /* End read_input and return to evaluation. */





/************  ordered_add_to_current_gen  **************************/
/* This function adds the latest simulation results to list of the 
    current generation's simulation results already processed. 
    Insertion is based on lexicographic ordering with the first 
    criterion arbitrarily chosen as the primary decision rule, 
    followed (in the case of ties on assess_res[1]) by the second 
    criteria, etc. The order allows us to calculate the Non-dominated 
    ranking of each simulation in one pass through the list. 
    For the tie_test:   
    tie =-1 -> latest ties comparison on all criteria 
    tie = 0 -> latest goes somewhere after comparison on list 
    tie = 1 -> latest goes before comparison on list                */  
/********************************************************************/

void ordered_add_to_current_gen(struct siminfo *latest_sim_ptr, 
            struct siminfo **ptr_to_head_location)
{
  int tie=0;
  struct siminfo *comparison_ptr;
  struct siminfo *previous_ptr;

  /* Assign the top of the lexicographically ordered list as the 
  comparison simulation. */

  comparison_ptr=*ptr_to_head_location; 
  previous_ptr=NULL;

  /* Test the latest simulation against the comparison simulation. */

  tie=tie_test(latest_sim_ptr, comparison_ptr,0);   

  /* If latest simulation does not tie or beat the comparison, keep 
  searching the list. Assign the comparison to the next simulation in 
  the lexicographically ordered list and test again. Repeat until 
  comparison is tied, beaten or the end of the list (=NULL) is reached. */

  while ( (comparison_ptr!=NULL) && tie==0) 
  {
   tie=tie_test(latest_sim_ptr, comparison_ptr,0);

   previous_ptr = comparison_ptr;
   comparison_ptr = comparison_ptr->next;
  }
 

  /* Link the latest simulation to the lexicographically ordered list 
  according to its tie test evaluation. If the simulation previous to 
  the comparison is NULL, then assign the latest to the top of the list 
  and return the pointer to the new top to evaluation. */ 

  
  latest_sim_ptr->next=comparison_ptr;

    if(previous_ptr == NULL)
    *ptr_to_head_location=latest_sim_ptr;
    else
     previous_ptr->next=latest_sim_ptr;

  }     /* End of the ordered_add_to_current_gen. */

 


/*********************** tie_test *******************************/
/* This function compares two simulation assessment vectors on 
    criteria, in original order of criteria listing, until a 
    difference is found. If BINARY, then the larger 'wins' a 
    comparison, if !BINARY then the smaller value 'wins' a comparison. 
    The function returns:
    -1 if tied on all criteria
     1 if latest is 'better' than comparison
     0 if latest is 'worse' than comparison                     */
/****************************************************************/

int tie_test(struct siminfo *latest_sim_ptr, struct siminfo *comparison_sim_ptr, int crit)
{
 int i,temp;


 /* Loop through all of the criteria */

  for(i=crit; i < NUMCRITERIA; i++)
    {

  /* Test whether the assessment criteria value are NOT equal between 
      the latest simulation and the comparison simulation until a 
      difference occurs. */

        if(latest_sim_ptr->assess_res.crit[i]!=
            comparison_sim_ptr->assess_res.crit[i])
        {
  /* If a difference occurs, compare them to see which criterion 
      wins. compare returns an integer of its evaluation. See compare 
      function for details. */

            temp=compare(latest_sim_ptr->assess_res.crit[i],
            comparison_sim_ptr->assess_res.crit[i]); 
            return(temp);
        }
    }
 
  /* If the simulations are equal on all criteria, a tie has occurred 
  and tie_test returns -1. */

  return (-1);

}   /* End of tie_test. */





/************************** compare  ****************************/
/* This function decides who wins in a comparison of two simulations
 at one specific criterion. Note: the decision rule depends
 on whether binary or continuous error functions are used 
 (coded in the constant BINARY).                                */
/****************************************************************/

int compare( double left_res, double right_res)
{

  /* If the search uses binary error measures and if the latest (left)
    simulation is larger than the comparison simulation (right), the 
    latest wins and compare returns 1 to tie_test. If the latest is 
    smaller than the comparison, it loses and compare returns 0 to tie_test. */

    if (BINARY==1)
    {
        if(left_res>right_res)
        { 
         return(1);
        }
        else
        {
         return(0);
        }
    }
  /* If the search uses continuous error measures and if the latest (left)
    simulation is smaller than the comparison simulation (right), the 
    latest wins and compare returns 1 to tie_test. If the latest is 
    larger than the comparison, it loses and compare returns 0 to tie_test. */

    else
    {
        if(left_res<right_res)
        {
         return(1);
        }
        else
        {
         return(0);
        }
    }       /* End of else. */
    
}   /* End of compare. */




/************************* assign_fitness *****************************/
/* This function passes through a lexicographically ordered list of the 
   current generation's simulations, figuring out each simulation's
   nondominated ranking (Pareto Ranking - see Srinivas & Deb, 1995),
   and their niche_count - i.e., a type of smoothed kernel density 
   estimator of the 'crowdedness' of a neighborhood around the simulation 
   in objective space. The niche_count assumes some sort of 
   commensurability among criteria error measures.
 
   For binary error measures, we have defined it as the sq. root of the 
   number of criteria not satisfied in common. For continuous measures, 
   it is assumed that assessed criteria results are all commensurable.
   If this is not true, the code needs to be changed to include some 
   sort of component-specific scaling factor in the distance measure.

   The function passes through the current_gen simulation list, 
   sequentially  selecting a current_sim(ulation) and comparing it to 
   every previous simulation (comparison_sim) on the list.

   At each comparison, the following calculations are made:

   1. Figure out the dominance relations among the two simulations
    and update each's nondominated ranking [non_dominated_ranking()].
   2. Calculate the 'distance' between simulations in objective/criteria
    space [criteria_distance()].
   3. Use this distance to update the niche-count of each simulation.
   4. Move to next comparison simulation, goto 1 unless next comparison
    = current. If so, set current=current->next and restart w/
    comparison at top of list.
   5. Pass down list and update max_niche_count for each Rank.
   6. Calculate baseline fitness for Ranks 1, 2, .... Baseline for
    Rank 1 = 100, Rank 2 = min(fitness Rank 1) = 100/max_niche_count(R1),
    similarly for other ranks. Assign final fitnesses, and store
    cumulative total in total_fitness.                     */
/***********************************************************************/

void assign_fitness(struct siminfo *current_gen_first_ptr, double *total_fitness_ptr)
{
  struct siminfo *current_sim;
  struct siminfo *comparison_sim;

  extern double max_niche_count[POP_SIZE];
  int i;


  /* Initialize the max_niche_counts for all the simulations (individuals) in the 
  current generation (population). */

   for (i=0; i<POP_SIZE; i++)
   {
       max_niche_count[i]=0.0;
   }
   
   /* Assign NDrank = 1 to the first simulation on the lexicographically ordered 
   list since it dominates on criterion 1. */

   current_gen_first_ptr->NDrank=1;

   /* Assign the next simulation on the list as the current simulation. */

   current_sim = current_gen_first_ptr->next;


   /* Start the nondominated ranking comparison with the second simulation, 
   the current simulation. Move through the simulations. */

   while (current_sim != NULL)
    {

    /* Initialize the nondominated ranking variable of the current simulation. */

     current_sim->NDrank = 0;

    /* Assign the first simulation in the lexicographically ordered list as 
     the comparison simulation. */

     comparison_sim=current_gen_first_ptr;


     /* Compare the current simulation against all of the simulations in the 
     lexicographically ordered list up to itself. */

     while(comparison_sim!=current_sim)
       {

       /* Figure out the nondominated ranking of the current simulation relative 
       to the comparison simulation. */

       non_dominated_ranking(current_sim,comparison_sim);


       /* Figure out the distance between the current simulation's assessment vector 
       (if binary) or criteria vector (if continuous) and the comparison simulation's 
       assessment vector or criteria vector in the criteria objective space. */

       criteria_distance(current_sim, comparison_sim);

       /* Move the comparison to the next simulation in the lexicographically 
       ordered list.  */

        comparison_sim = comparison_sim->next; 
       }
       /* Move the current to the next simulation in the list.  */

     current_sim=current_sim->next;
    }


    /* After all ranks are assigned, find the max_niche_count for a given rank. There 
    may be as many different ranks as the size of the generation (population= POP_SIZE). 
    Start at the top of the list. Figure out each simulation's rank and update for the 
    maximum of its rank. */

    current_sim = current_gen_first_ptr;

     while (current_sim != NULL)
        {
        max_niche_count[current_sim->NDrank -1] =
            max(max_niche_count[current_sim->NDrank -1],current_sim->niche_count);

        current_sim=current_sim->next;

        }       /* End max_niche_count update. */


   /* Calculate each rank's baseline fitness, then downweight the rank by its 
   max_niche_count. Assign fitness measures to each simulation based on its rank, 
   cumulating into a total_fitness measure for the generation. */

    share_fitness(current_gen_first_ptr, max_niche_count, total_fitness_ptr);

}   /* End of assign_fitness. */




/* ***************** non_dominated_ranking ***********************/
/* This function calculates nondominated ranking (pareto ranking) 
    of each simulation relative to current generation's results. */
/*****************************************************************/

void non_dominated_ranking(struct siminfo *current_sim,
        struct siminfo *comparison_sim)
{
    int i, dom_index=0;
 
    /* Loop through all the criteria, examining whether the current simulation 
    "Pareto" dominates the comparison simulation. Sum up the counts of domination 
    criteria.   */

    for (i=0; i<NUMCRITERIA; i++)
    {
        dom_index+= crit_comp(current_sim->assess_res.crit[i],
              comparison_sim->assess_res.crit[i]);
    }

    /* Assign the comparison simulation's rank to the current  simulation if 
    their assessment results are equivalent. */

    if (dom_index==0) 
  { 
   current_sim->NDrank = comparison_sim->NDrank;
  }

    /* Assign the comparison simulation's rank plus one or the current simulation's 
    rank, whichever is larger, to the current simulation, if the comparison 
    simulation's assessment dominates the current simulation's assessment. */

 else if (dom_index<0)
  { 
   current_sim->NDrank = max(comparison_sim->NDrank+1,current_sim->NDrank);
  }
 else           
 
 /* If the current simulation is codominant with the comparison, assign the 
 current simulation the same rank as the comparison simulation's rank if the 
 current simulation has not yet received a rank. */

  { 
   if (current_sim->NDrank==0) current_sim->NDrank = comparison_sim->NDrank;
  }

}   /* End of non_dominated_ranking. */



/******************************** crit_comp ******************************/
/* This function compares the assessment criteria of the current simulation 
(Left) with that of the comparison simulation (Right), one criterion at a time. 
If the search is with BINARY error measures:
    if the current equals the comparison, the function returns 0,
    if the current is greater than the comparison, the function returns 
    integer=NUMCRITERIA,
    if the current is less than the comparison, the function returns -1.

If the search is with continuous error measures:
    if the current equals the comparison, the function returns 0,
    if the current is less than the comparison, the function returns 
    integer=NUMCRITERIA,
    if the current is greater than the comparison, the function returns -1.     */
/****************************************************************/

int crit_comp(double Left, double Right)
{
    /* The criteria are equal. */

    if (Left==Right) return(0);         
 


    /* The search is with binary error measures. */

    else if (BINARY)
     {
     /* The current (Left) dominates the comparison (Right). */

       if (Left>Right)      return(NUMCRITERIA);
       
     /* The current (Left) is dominated by the comparison (Right). */

       else                 return(-1);
     }
 


    /* The search is with continuous error measures. */

    else
     {

     /* The current (Left) dominates the comparison (Right). */

      if (Left<Right)       return(NUMCRITERIA);

     /* The current (Left) is dominated by the comparison (Right). */

      else                  return(-1);
     }

}   /* End of crit_comp. */





/*************************** criteria_distance ***************************/
/* This function calculates the distance between two simulations in 
   the criteria (objective) space. It is used in calculating each 
   simulation's niche_count, which is used to 'share fitness' across 
   neighborhoods so as to combat genetic drift and force the optimization 
   search to explore the full Pareto Front.

 NOTE: WE HAVE MADE THE FOLLOWING ASSUMPTIONS IN ORDER TO DO THIS:
 BINARY: distance = euclidean distance between the assessment vectors
         as pts in the NUMCRITERIA-dimensional unit hypercube.
 CONTINOUS: distance = euclidean distance among critera_res vectors,
         assuming each criterion result is on the same scale (e.g., 
         [0,1]).
 See Srinivas & Deb. 1995. Evolutionary Computation 2(3):221-248. for
 lead into fitness sharing literature.                                  */
/************************************************************************/

void criteria_distance(struct siminfo *current_sim, struct siminfo *comparison_sim)
{
    double dist=0.0;
    int i;
 

    /* Loop through all of the criteria. */

    for (i=0; i<NUMCRITERIA; i++)
      {
    /* Square the difference between the current simulation and the 
    comparison simulation, then sum the squared differences. */

       if (BINARY) dist+=SQR(current_sim->assess_res.crit[i] 
                - comparison_sim->assess_res.crit[i]);
       else dist+=SQR(current_sim->cresults.crit[i] - 
              comparison_sim->cresults.crit[i]);
      }
 
    /* Downweight the niche counts of the current simulation and the comparison 
    simulation, depending on the distance between the two simulations in the 
    criteria objective space and the maximum sharing distance (MAX_SHARE_DIST). */

    current_sim->niche_count += sharing(sqrt(dist));
    comparison_sim->niche_count +=sharing(sqrt(dist));

}   /* End of criteria_distance. */




/******************************* sharing ***********************************/
/* This function applies a sharing function to downweight the distance 
calculations to derive the cumulative neighborhood niche counts between the 
current simulation and the comparison simulation. 
From pg. 228 in Srinivas & Deb 1995 Evolutionary Computation 2(3): 221-248.*/
/***************************************************************************/

double sharing(double dist)
{
   /* If the distance between the simulations is less than or equal to the 
    maximum sharing distance, the function returns a downweight of the distance. */

   if (dist <=MAX_SHARE_DIST)    return(1 - SQR(dist/MAX_SHARE_DIST));


   /* If the distance is greater than the maximum sharing distance, the 
   function returns 0.0, impling the simulations are too far apart to share fitness. */

   else                          return(0.0);

}   /* End of sharing.  */





/***************************** share_fitness *******************************/
/* This function assigns each rank a baseline fitness. This baseline fitness 
    is then downweighted ('shared') by the rank's max_niche_count. That is, 
    downweight the baseline fitness for the ranks of those simulations which 
    are near each other in objective space.    */
/***************************************************************************/

void share_fitness(struct siminfo *current_gen_first_ptr,
        double max_niche_count[POP_SIZE], double *total_fitness_ptr)
{
    int i;
    double baseline[POP_SIZE];
    struct siminfo *current_sim;
    double total_fit;

    total_fit = *total_fitness_ptr;



    /* Initialize the baseline fitness for each rank. There may be as many 
    ranks as the population size. */

    for (i=0; i<POP_SIZE; i++)
    {
        baseline[i]=0.0;
    }


    /* Assign the highest baseline fitness to the top rank level (NDRank=1). */

    baseline[0]=100.0;  

    /* Calculate each rank's baseline fitnesss downweighted by its max_niche_count.
    The i indicates rank. Loop through all possible ranks. */

    for (i=1; i<POP_SIZE; i++)
     { 
      baseline[i]=(baseline[i-1]*WGT)*(1/(max_niche_count[i-1]) );
     }
   

    /* Assign the current simulation to the top of the lexicographically 
    ordered list of the current generation's simulations.  */

    current_sim=current_gen_first_ptr;  
 
    /* Calculate each simulation's actual fitnesss for its rank downweighted 
    by its actual niche_count. Sum up each simulation's fitness to find the 
    total fitness measure for the current generation. */
   
    while( current_sim != NULL)
     {
      current_sim->fitness = baseline[current_sim->NDrank -1]/
        current_sim->niche_count;
      total_fit += current_sim->fitness;
      current_sim = current_sim->next;
     }



   /* Return the total fitness to the call function, assign_fitness.  */

    *total_fitness_ptr=total_fit;

}   /* End of share_fitness. */





/*********************** Pareto_Frontier_update **************************/
/* This function moves through the lexicogrpahically ordered list, comparing 
    each nondominated rank 1 simulation to the 'Historical Pareto Frontier'
    and updating the Frontier as necessary. The 'Historical Pareto Frontier' 
    refers to those simulations that, out of all the simulations investigated 
    in all generations so far, are nondominated. Simulations in the Frontier 
    are referred to as members.                                          */
/*************************************************************************/

void Pareto_Frontier_update(struct siminfo *current_gen_first_ptr,
    struct pset **ptr_to_Frontier_first_location,int generation)
{
    struct siminfo *current_sim;
    struct pset *pareto_member;
    int i, on_list, stop_loop;
    struct pset *new_pareto_member;


    /* If the Historical Pareto Frontier has NOT been initalized, assign the 
    first simulation in the lexicogrpahically ordered list as the first member 
    (current) in the Frontier. */

    if(*ptr_to_Frontier_first_location == NULL)
     {
        current_sim = current_gen_first_ptr;

    /* Allocate memory for the Frontier's first member. */

        *ptr_to_Frontier_first_location = 
            (struct pset *)malloc(sizeof(struct pset));
        pareto_member = *ptr_to_Frontier_first_location;
        pareto_member->next = NULL;


    /* Copy the first simulation's assessment vector into the Frontier's 
    first member. */

        for(i=0; i < NUMCRITERIA; i++)
        {
            pareto_member->assess.crit[i] = current_sim->assess_res.crit[i];
        }

    /* Allocate memory for the first simulation's parameterization information 
    and copy information into the Frontier's first member.  */

            pareto_member->psim = 
            (struct psiminfo *) malloc (sizeof(struct psiminfo));

        (pareto_member->psim)->next = NULL;

        for(i=0; i < NUMPARAMS; i++)
          {
            (pareto_member->psim)->paramvalues.par[i]= current_sim->paramvalues.par[i];
          }


    /* Copy the first simulation's criteria vector into the Frontier's first 
    member. */

        for(i=0; i < NUMCRITERIA; i++)
          {
        (pareto_member->psim)->cresults.crit[i]= current_sim->cresults.crit[i];
          }

    /* Increment the investigation to the next simulation in the lexico- 
    graphically ordered list for the current generation to start the comparison
    to the Historical Pareto Frontier with the second simulation in the current 
    generation. */

        current_sim = current_gen_first_ptr->next;  


     }  /* End of initialization of Frontier with the current_gen_first_ptr.  */


    /* If the Frontier was already initialized, start at the first simulation, 
    at the top of the lexicographically ordered list, for the current generation. */

    else
    {
        current_sim = current_gen_first_ptr;
     }



    /* Loop through the entire list of simulations for the current generation. */

    while(current_sim != NULL)
     {

        /* Examine only the nondominated simulations in the current generation. */
        
        if (current_sim->NDrank==1)
        {
        
        /* Initialize the on_list and stop_loop flags.   */

            on_list = 0;      
            stop_loop=0;
 
        /* Start the comparison of the current simulation to the first member 
        in the  Historical Pareto Frontier.   */
            
            pareto_member=*ptr_to_Frontier_first_location;
            while ((pareto_member!=NULL) && (stop_loop == 0))
            {

        /* Compare the current simulation to the Frontier's first member. 
        If the current simulation's assessment vector is already in the 
        Frontier, Pareto_Comparison adds its parameterization and criteria results to the 
        Frontier, then exits the while loop.  */
        /* If a current simulation dominates the Frontier, it will replace 
        the pareto_member and the while loop will continue down the Frontier list, 
        comparing the new pareto_member with the other pareto_members to see if 
        they are also dominated or are codominant with the new pareto_member. */

                Pareto_Comparison(current_sim, &pareto_member, 
                    ptr_to_Frontier_first_location,&on_list, &stop_loop);

        /* Otherwise increment the investigation to the next member in the Frontier 
        and return to the while evalution. */

                pareto_member = pareto_member->next;
            }       

 /* End comparison against Pareto Frontier. */
        

 /* At this point, the current simulation has been added to the Frontier 
    (stop_loop = 1),is dominated by the Historical Pareto Frontier (stop_loop =1) 
    or it is codominant with a member of the Frontier (stop_loop=0) and 
    (on_list = 0). If it is codominant, the current simulation must be added 
    to the Frontier, at the end of the Frontier linked list.    */

            if ((stop_loop == 0) && (on_list == 0)) 
            {
                
        /* Start with the first member, at the top of the Historical 
        Pareto Frontier linked list. */

                pareto_member = *ptr_to_Frontier_first_location;


        /*  Loop through the Frontier's linked list to find the end. */

                while (pareto_member->next != NULL)     
                    pareto_member = pareto_member->next;


        /* Allocate memory for a new pareto member and copy the current 
        simulation's assessment vector into it.  */

                new_pareto_member=(struct pset *)malloc(sizeof(struct pset));

                new_pareto_member->next = NULL;

                for(i=0; i < NUMCRITERIA; i++)
                {
                    new_pareto_member->assess.crit[i] = 
                        current_sim->assess_res.crit[i];
                }


        /* Allocate memory for the new pareto member's parameterization 
        and criteria vector and copy the current simulation's 
        parameterization and criteria vector into it.  */

                new_pareto_member->psim = 
                    (struct psiminfo *) malloc (sizeof(struct psiminfo));
    
                (new_pareto_member->psim)->next = NULL;

                for(i=0; i < NUMPARAMS; i++)
                {
                    (new_pareto_member->psim)->paramvalues.par[i]= 
                        current_sim->paramvalues.par[i];
                }

                for(i=0; i < NUMCRITERIA; i++)
                {
                  (new_pareto_member->psim)->cresults.crit[i]= 
                        current_sim->cresults.crit[i];
                }


        /* Link the last pareto member to the new pareto member. */

                pareto_member->next = new_pareto_member;
            } /* End adding current to end of list. */
        }         /* done with this ND rank 1 simulation. */ 

        /* Increment to the next nondominated simulation in the current 
        generation and repeat the comparison process against the 
        Historical Pareto Frontier. */

      current_sim = current_sim->next;


     }      

}   /* End of Pareto_Frontier_update. */






/**************************** Pareto_Comparison ***************************/
/* This function compares two simulations, decides dominance and updates the 
    Historical Pareto Frontier as necessary from dominance results.     

  The Pareto comparison results in several outcomes. 
    If the dom_index = 0, it indicates a tie on all criteria. 
    If the dom_index < 0, it indicates the Pareto Frontier dominates the
        current simulation. 
    If dom_index >0 and NUMCRITERIA = 0, it indicates the current simulation 
        dominates the Pareto Frontier. 
    If dom_index >0 and NUMCRITERIA > 0, it indicates the current simulation 
       and the Pareto Frontier are codominant.                            */ 
/**************************************************************************/

void Pareto_Comparison( struct siminfo *current_sim, struct pset **ptr_to_pareto_member,
        struct pset **ptr_to_Frontier_first_location, int *on_list, 
        int *stop_loop)
{
    int i, dom_index=0;
    struct pset *temp_pareto_ptr;
    struct psiminfo *new_sim_ptr;
    struct pset *local_pareto_member;


    local_pareto_member=*ptr_to_pareto_member;

    /* Compare the assessment criteria results of the current simulation and 
    the comparison simulation, a member of the Historical Pareto Frontier, 
    one criterion at a time. Sum the criterion comparison results over all 
    of the criteria. */

    for (i=0; i<NUMCRITERIA; i++)
     { 
       dom_index += crit_comp(current_sim->assess_res.crit[i],
        local_pareto_member->assess.crit[i]);

     }
 
    /* If the current simulation and the comparison simulation have equivalent 
    assessment results, add the current simulation as a new element to the 
    comparison simulation's (Frontier member) element list. An element list 
    includes all of the parameterizations and their corresponding criteria 
    vectors that have the same assessment vector result. An element in the list 
    is a single parameterization and its corresonding criteria vector. Equivalent 
    assessment results usually only occurs in binary error measure searches. */

    if (dom_index == 0)  
     {
    
    /* Allocate memory for adding a new element to the Frontier member's 
    element list. */ 

       new_sim_ptr = (struct psiminfo *) malloc (sizeof (struct psiminfo)); 

    /* Copy the current simulation's parameterization into the new element. */

       for(i=0; i < NUMPARAMS; i++)
          {
            new_sim_ptr->paramvalues.par[i]= current_sim->paramvalues.par[i];
          }

     /* Copy the current simulation's criteria vector into the new element. */

       for(i=0; i < NUMCRITERIA; i++)
          {
            new_sim_ptr->cresults.crit[i]= current_sim->cresults.crit[i];
          }


  /* Link the new element to the top of the Frontier member's element list, 
    then send flag to stop comparing the current simulation to other members 
    of the Frontier. Return to calling function, Pareto_Frontier_update. */

       new_sim_ptr->next = local_pareto_member->psim;
       local_pareto_member->psim = new_sim_ptr;
       *stop_loop = 1;

       return;
     }      /* End of tie situation. */


    /* If the comparison simulation dominates the current simulation in terms 
    of the assessment vector, send flag to stop comparing the current simulation 
    to the Historical Pareto Frontier. Return to calling function, 
    Pareto_Frontier_update. */

    if (dom_index <0)   
    {
      *stop_loop=1;     
       return;
     }      /* End of domination situation. */

    /* If the current simulation dominates the comparison simulation in terms 
    of the assessment vector, replace the dominated comparison simulation by 
    the current simulation. */


    if ((dom_index>0) && ( (dom_index % NUMCRITERIA) == 0))
     {  


    /* Check to see if current simulation has NOT already been added to the 
    Frontier. If the current simulation is NOT in the Frontier, on_list = 0. */

     if (*on_list == 0)
       {    
    
    /* Copy the current simulation's assessment results into the dominated 
    comparison simulation's memory space in the Frontier linked list 
    (pareto_member->assess). */
  
        for(i=0; i < NUMCRITERIA; i++)
            {
              local_pareto_member->assess.crit[i] = current_sim->assess_res.crit[i];
            }
   
    /* Copy the current simulation's parameterization and criteria vector into 
    the top of the dominated comparison simulation's element list.*/

        for(i=0; i < NUMPARAMS; i++)
          {
            (local_pareto_member->psim)->paramvalues.par[i]= 
                current_sim->paramvalues.par[i];
          }

         for(i=0; i < NUMCRITERIA; i++)
            {
            (local_pareto_member->psim)->cresults.crit[i]= 
                current_sim->cresults.crit[i];
            }


    /* Delete (free up memory of) any additional elements in the dominated 
    comparison simulation's element list. Recall that the element list included 
    all the parameterizations and their corresponding criteria vectors that 
    had the same assessment vector as the dominated comparison simulation. */
         

        if( (local_pareto_member->psim)->next != NULL)
        {
            /* delete all following psims on list*/
            free_rest_list((local_pareto_member->psim)->next);
            
            /* now set ptr to NULL */
            (local_pareto_member->psim)->next= NULL;
        }
     
        /* Next command seems unnecessary information 
        if(local_pareto_member->next !=NULL)
        {
            printf("loc_pareto_member->next->assess.crit[0]=%d\n",
            (local_pareto_member->next)->assess.crit[0]);
        } */

    /* Send flag indicating that the current simulation is in the Frontier. */

        *on_list = 1;   

    /* Return to calling function, Pareto_Frontier_update. */ 

        return;     
       }        /* End adding to list situation. */


    else 

    /* After the current simulation dominates a comparison simulation from the 
    Frontier member and replaces that simulation in the Frontier, compare the 
    current simulation to the rest of the Frontier. If it dominates other members 
    of the Frontier, delete the other members. */
        
       {

    /* Point to the top of Frontier and move through linked list until the pointer, 
    temp_pareto_ptr, points to the current simulation that dominated and replaced a 
    Frontier member. */
        
        temp_pareto_ptr=*ptr_to_Frontier_first_location;
        while (temp_pareto_ptr->next != local_pareto_member) 
            temp_pareto_ptr = temp_pareto_ptr->next;


    /* Once the dominating current simulation is being pointed to by the 
    temp_pareto_ptr, link the dominating current simulation to local_pareto_member->next, 
    the member that follows the dominated member, local_pareto_member. */
  
        temp_pareto_ptr->next = local_pareto_member->next; 
    
    /* Free up the local_pareto_member space. */

        local_pareto_member->next=NULL;
        free_pareto(local_pareto_member);

    /* Reset the pareto_member address to continue the comparison loop in the
    Pareto_Frontier_update function. */

        *ptr_to_pareto_member=temp_pareto_ptr;


    /* Return to calling function, Pareto_Frontier_update. */ 

        return; 
       } 
    /* End of the current dominates situation.  */

     }  
     
    /*Else the current simulation and the comparison simulation are codominant, 
    so keep looking down the Historical Pareto Frontier list. */


}   /* End of the Pareto_Comparison. */





/******************* free_rest_list ********************************/
/* This function runs down a psiminfo list and frees all the memory. 
It is called by the Pareto_Comparison function.                    */
/*******************************************************************/

void free_rest_list(struct psiminfo *start_simptr)
{
 struct psiminfo *simptr;


 if (start_simptr!=NULL)            /* something to delete */   
  {
  while (start_simptr->next!=NULL)  /* more than this element to delete */
   {
    simptr=start_simptr;            /* start at beginning of 'deletion' list */
     while (simptr->next->next!=NULL)
       simptr=simptr->next;         /* find ptr to last element */
    free(simptr->next);             /* free last element */
    simptr->next = NULL;            /* make new end */
   }                                /* while: repeat process */

  free(start_simptr);               /* delete first element */
  }                                 /* end if*/
}                                   /* end free_rest_list */





/******************* free_rest_sim_list ****************************/
/* This function runs down a siminfo list and frees all the memory. 
It is called by the main function.                                 */
/*******************************************************************/

void free_rest_sim_list(struct siminfo *start_simptr)
{
 struct siminfo *simptr;

 if (start_simptr!=NULL)            /* something to delete */
  {
  while (start_simptr->next!=NULL)  /* more than this element to delete */
   {
    simptr=start_simptr;            /* start at beginning of 'deletion' list */
     while (simptr->next->next!=NULL)
       simptr=simptr->next;         /* find ptr to last element */
    free(simptr->next);             /* free last element */
    simptr->next = NULL;            /* make new end */
   }                                /* while: repeat process */

  free(start_simptr);               /* delete first element */
  }                                 /* end if*/
}                                   /* end free_rest_list */




/************************  free_pareto  ****************************/
/* This function frees up the memory of the Pareto List. It is called
    by the main function.                                          */
/*******************************************************************/

void free_pareto(struct pset *start_pset)
{
   struct pset *temp;

   if (start_pset != NULL)
    {
     while (start_pset->next != NULL)       /* more elements to delete*/
     {
      temp=start_pset;                      /*find ptr to last element */
      while(temp->next->next!=NULL) temp=temp->next; 
      free_rest_list(temp->next->psim);     /* remove sim list */
      free(temp->next);
      temp->next = NULL;
     }                                      /* while:  repeat process */
     free_rest_list(start_pset->psim);
     free(start_pset);
    }
}                                           /* end free_pareto */






 
 
